(*+--------------------------------------------------------------------------+
  |                                                                          |
  |                           Analyseur micro-caml                           |
  |                                                                          |
  +--------------------------------------------------------------------------+*)


type mot =
  | Ident   of string       (* identificateur *)
  | Nombre  of int          (* constante      *)
  | Opr     of int * string (* opr. infixe    *)
  | Par     of string       (* parenthse     *)
;;

     (*+-------------------------------------------------------+
       |  Transforme une liste de caractres en liste de mots  |
       +-------------------------------------------------------+*)

let mots =

  let rec suite(chaine,i) = match chaine.[i] with
  | `$`                           -> []
  | (` `|`\t`|`\n`)               -> suite(chaine,i+1)
  | (`A`..`Z`|`a`..`z`)           -> ident(chaine,i,i+1)
  | (`0`..`9` as c)               -> nombre(chaine,int_of_char(c)-48,i+1)
  | (`*`|`/`)                     -> opr(3,chaine,i,i)
  | (`+`|`-`)                     -> opr(2,chaine,i,i)
  | (`<`|`>`|`=`)                 -> opr(1,chaine,i,i)
  | (`(`|`)` as c)                -> Par(make_string 1 c) :: suite(chaine,i+1)
  | c -> failwith ("caractre non reconnu : " ^ (make_string 1 c))

  and ident(chaine,i,j) = match chaine.[j] with
  | (`A`..`Z`|`a`..`z`|`_`|`0`..`9`) -> ident(chaine,i,j+1)
  | _ -> Ident(sub_string chaine i (j-i)) :: suite(chaine,j)

  and nombre(chaine,v,j) = match chaine.[j] with
  | (`0`..`9` as c) -> nombre(chaine,10*v+int_of_char(c)-48,j+1)
  | _ -> Nombre v :: suite(chaine,j)

  and opr(p,chaine,i,j) = match chaine.[j] with
  | (`+`|`-`|`*`|`/`|`<`|`>`|`=`) -> opr(p,chaine,i,j+1)
  | _ -> Opr(p,sub_string chaine i (j-i)) :: suite(chaine,j)

  in fun chaine -> suite(chaine^"$",0)
;;


                        (*+-----------------+
                          |  Mise en arbre  |
                          +-----------------+*)



type expression = 
  | Const of int                                  (* nombre              *)
  | Nom   of string                               (* identificateur      *)
  | Fun   of string * expression                  (* fonction d'une var. *)
  | Bin   of expression * string * expression     (* oprateur binaire   *)
  | Appl  of expression * expression              (* appl. de fonction   *)
  | If    of expression * expression * expression (* slection           *)
  | Letin of string * expression * expression     (* dfinition locale   *)
;;

(* pile hybride contenant des expressions et des tokens non encore rduits *)
type phrase_lment =
  | E of expression                                 (* expression *)
  | Def of string * expression                      (* dfinition *)
  | OPR of int * string                             (* op binaire *)
  | LET | EGAL | IN | IF | THEN | ELSE | PAR        (* dlimiteur *)
;;

exception Stxerr of phrase_lment list;;

let arbre =

  (* empile une expression *)
  let rec push exp pile = match pile with
  | E(f) :: suite -> push (Appl(f,exp)) suite
  | _ -> E(exp) :: pile

  (* effectue les oprations de priorit >= p *)
  and opr(p,pile) = match pile with
  | E(b) :: OPR(q,o) :: E(a) :: suite when q >= p
    			-> opr(p, push (Bin(a,o,b)) suite)
  | _ -> pile

  (* rduction des groupes if-then-else et let *)
  and rduit(pile) = match opr(0,pile) with
  | E(c) :: ELSE :: E(b) :: THEN :: E(a) :: IF :: suite
    			-> rduit(push (If(a,b,c)) suite)
  | E(e) :: EGAL :: E(Appl(f,Nom(x))) :: LET :: suite
    			-> rduit(E(Fun(x,e)) :: EGAL :: E(f) :: LET :: suite)
  | E(e) :: IN :: E(y) :: EGAL :: E(Nom(x)) :: LET :: suite
    			-> rduit(push (Letin(x,y,e)) suite)
  | p -> p

  (* parenthse fermante *)
  and parf(pile) = match rduit(pile) with
  | E(a) :: PAR :: suite ->  push a suite
  | p -> raise(Stxerr p)
  in

  (* traitement d'un lexme *)
  let rec job pile t = match t with
  | Ident("let")  -> LET  :: pile
  | Ident("if")   -> IF   :: pile
  | Ident("in")   -> IN   :: rduit(pile)
  | Ident("then") -> THEN :: rduit(pile)
  | Ident("else") -> ELSE :: rduit(pile)
  | Opr(p,"=")    -> (match pile with E(_) :: LET :: _ -> EGAL :: pile
                                    | _ -> OPR(p,"=") :: opr(p,pile))
  | Opr(p,o)      -> OPR(p,o) :: opr(p,pile)
  | Ident(i)      -> push (Nom(i))   pile
  | Nombre(n)     -> push (Const(n)) pile
  | Par("(")      -> PAR :: pile
  | Par(")")      -> parf(pile)
  | _             -> failwith "token imprvu"

  in fun liste -> match rduit(it_list job [] liste) with
  | [E(a)] -> E(a)
  | [E(e); EGAL; E(Nom(x)); LET] -> Def(x,e)
  | p -> raise(Stxerr p)

;;

                  (*+-----------------------------+
                    |  Impression semi-graphique  |
                    +-----------------------------+*)

#open "arbres";;
load_object "arbres";;

let rec garbre(exp) = match exp with
| Const(i)         -> G_noeud(string_of_int(i),[])
| Nom(x)           -> G_noeud(x,[])
| Fun(x,e)         -> G_noeud("fun", [G_noeud(x,[]); garbre(e)])
| Bin(a,o,b)       -> G_noeud(o, [garbre(a); garbre(b)])
| Appl(f,x)        -> G_noeud("appl", [garbre(f); garbre(x)])
| If(c,v,f)        -> G_noeud("if", [garbre(c); garbre(v); garbre(f)])
| Letin(x,e,f)     -> G_noeud("letin", [G_noeud(x,[]); garbre(e); garbre(f)])
;;

let print_exp(e) = print_string_g_arbre(garbre(e));;
install_printer "print_exp";;
#close "arbres";;
