(*+--------------------------------------------------------------------------+
  |                                                                          |
  |                         Compilation d'une expression                     |
  |                                                                          |
  +--------------------------------------------------------------------------+*)

type donne = Int of int | Ptr of int;; (* nombre ou pointeur *)

type instruction =
  | Load of adresse * donne                     (* chargement immdiat   *)
  | Move of adresse * adresse                    (* dplacement de donne *)
  | Push of adresse                              (* empilement            *)
  | Pop  of adresse                              (* dpilement            *)
  | Op   of string * adresse * adresse * adresse (* opr, dest, op1, op2  *)
  | Add  of reg * int                            (* incrmente registre   *)
  | Call of adresse                              (* appel de fonction     *)
  | Ret
  | Jfalse of adresse * int                      (* branchement si faux   *)
  | Jptr   of adresse * int                      (* branch. si pointeur   *)
  | Jmp    of int                                (* br. inconditionnel    *)
  | Print  of adresse
  | Comm   of string                             (* commentaire           *)
  | Halt
and adresse =
  | Mem of reg * int (* adr. indirecte : registre, dpl. *)
  | Reg of reg       (* no de registre                   *)
and reg =
  | PC (* compteur de programme  *)
  | SP (* pointeur de pile       *)
  | EP (* pointeur environnement *)
  | HP (* pointeur sommet du tas *)
  | DR (* registre de donnes    *)
;;

                         (*+---------------+
                           |  compilation  |
                           +---------------+*)

(* envir = liste des symboles          *)
(* code  = mmoire de programme        *)
(* pc    = index de "code" (rfrence) *)
(* exp   = expression  compiler       *)
(* le code compil doit laisser le rsultat au sommet de la pile *)
let rec comp_ex envir code pc exp = match exp with

(* constante : l'empile *)
| Const(c) ->
    code.(!pc) <- Add(SP,1);              incr pc;
    code.(!pc) <- Load(Mem(SP,0),Int(c)); incr pc

(* identificateur : le cherche dans l'environnement et empile la valeur *)
| Nom(x) ->
    let rang = index x envir in
    code.(!pc) <- Move(Reg(DR),Reg(EP));     incr pc;
    for i=1 to rang do
      code.(!pc) <- Move(Reg(DR),Mem(DR,1)); incr pc;
    done;
    code.(!pc) <- Push(Mem(DR,0));           incr pc

(* opration binaire : l'excute sur la pile *)
| Bin(a,o,b) ->
   comp_ex envir code pc a;
   comp_ex envir code pc b;
   code.(!pc) <- Op(o,Mem(SP,-1),Mem(SP,-1),Mem(SP,0)); incr pc;
   code.(!pc) <- Add(SP,-1);                            incr pc

(* test : l'excute *)
| If(c,v,f) ->
   comp_ex envir code pc c;
   code.(!pc) <- Pop(Reg(DR)); incr pc;  (* cond. -> DR *)
   let p0 = !pc in incr pc; (* branchement  dterminer *)
   comp_ex envir code pc v;
   let p1 = !pc in incr pc; (* branchement  dterminer *)
   code.(p0) <- Jfalse(Reg(DR),!pc);
   comp_ex envir code pc f;
   code.(p1) <- Jmp(!pc)

(* fonction : compile le code et cre une clture *)
| Fun(x,e) ->
   let p0 = !pc in incr pc; (* branchement  dterminer *)

   (* compile l'expression *)
   comp_ex (x :: envir) code pc e;
   code.(!pc) <- Pop(Reg(DR)); incr pc; (* rs -> DR *)
   code.(!pc) <- Ret;          incr pc;
   code.(p0)  <- Jmp(!pc);

   (* alloue une paire dans le tas et retourne son adresse *)
   code.(!pc) <- Add(HP,-2);                 incr pc;
   code.(!pc) <- Load(Mem(HP,0),Ptr(p0+1));  incr pc;
   code.(!pc) <- Move(Mem(HP,1),Reg(EP));    incr pc;
   code.(!pc) <- Push(Reg(HP));              incr pc;

(* appl : change d'environnement et appelle le S/P *)
| Appl(f,x) ->

   (* sauvegarde le pointeur de tas et calcule f et x *)
   code.(!pc) <- Push(Reg(HP));              incr pc;
   comp_ex envir code pc x;
   comp_ex envir code pc f;
   
   (* recopie la valeur de x dans le tas et la lie  l'environnement de f *)
   code.(!pc) <- Add(HP,-2);                 incr pc;
   code.(!pc) <- Move(Mem(HP,0),Mem(SP,-1)); incr pc;
   code.(!pc) <- Move(Reg(DR),Mem(SP,0));    incr pc;
   code.(!pc) <- Move(Mem(HP,1),Mem(DR,1));  incr pc;
   code.(!pc) <- Add(SP,-2);                 incr pc;
   
   (* appelle f avec le nouvel environnement (rs. dans DR) *)
   code.(!pc) <- Push(Reg(EP));              incr pc;
   code.(!pc) <- Move(Reg(EP),Reg(HP));      incr pc;
   code.(!pc) <- Call(Mem(DR,0));            incr pc;
   code.(!pc) <- Pop(Reg(EP));               incr pc;
   
   (* si le rsultat est un entier, libre le tas *)
   code.(!pc) <- Jptr(Reg(DR), !pc+2);       incr pc;
   code.(!pc) <- Move(Reg(HP),Mem(SP,0));    incr pc;
   
   (* rs -> pile *)
   code.(!pc) <- Move(Mem(SP,0),Reg(DR));    incr pc;

(* letin : cre une entre pour x puis calcule e et f *)
| Letin(x,e,f) ->

    (* sauve HP et rserve une entre pour x dans l'environnement *)
    code.(!pc) <- Push(Reg(HP));              incr pc;
    code.(!pc) <- Add(HP,-2);                 incr pc;
    code.(!pc) <- Move(Mem(HP,1),Reg(EP));    incr pc;
    code.(!pc) <- Move(Reg(EP),Reg(HP));      incr pc;

    (* calcule e et affecte le rsultat  x *)
    comp_ex (x :: envir) code pc e;
    code.(!pc) <- Pop(Mem(EP,0));             incr pc;

    (* calcule f et restaure l'environnement *)
    comp_ex (x :: envir) code pc f;
    code.(!pc) <- Move(Reg(EP),Mem(EP,1));    incr pc;
    code.(!pc) <- Jptr(Mem(SP,0), !pc+2);     incr pc;
    code.(!pc) <- Move(Reg(HP),Mem(SP,-1));   incr pc;
    code.(!pc) <- Pop(Mem(SP,-1));            incr pc
;;

(* compile une phrase *)

let envir = ref []
and code  = make_vect 1000 Halt
and pc    = ref 0;;

let compile texte = match arbre(mots texte) with

(* expression : value, affiche puis nettoie le tas *)
| E(e) ->
    code.(!pc) <- Comm(texte);      incr pc;
    code.(!pc) <- Push(Reg(HP));    incr pc;
    comp_ex !envir code pc e;
    code.(!pc) <- Print(Mem(SP,0)); incr pc;
    code.(!pc) <- Add(SP,-1);       incr pc;
    code.(!pc) <- Pop(Reg(HP));     incr pc

(* dfinition : rserve une entre pour x puis calcule e *)
| Def(x,e) ->
    code.(!pc) <- Comm(texte);               incr pc;
    code.(!pc) <- Add(HP,-2);                incr pc;
    code.(!pc) <- Move(Mem(HP,1),Reg(EP));   incr pc;
    code.(!pc) <- Move(Reg(EP),Reg(HP));     incr pc;
    envir := x :: !envir;
 
    code.(!pc) <- Push(Reg(HP));             incr pc;
    comp_ex !envir code pc e; 
    code.(!pc) <- Jptr(Mem(SP,0), !pc+2);    incr pc;
    code.(!pc) <- Move(Reg(HP),Mem(SP,-1));  incr pc;
    code.(!pc) <- Move(Mem(EP,0),Mem(SP,0)); incr pc;
    code.(!pc) <- Add(SP,-2);                incr pc;

    code.(!pc) <- Print(Mem(EP,0));          incr pc

(* autres cas impossible *)
| _ -> failwith "cas impossible"
;;
    
         
                        (*+-----------------+
                          |  Dsassembleur  |
                          +-----------------+*)

#open "printf";;

let dsassemble ch a1 a2 =

  let pr fmt = fprintf ch fmt in

  let preg r = pr "%s" (match r with
  | PC -> "PC"
  | SP -> "SP"
  | EP -> "EP"
  | HP -> "HP"
  | DR -> "DR") in

  let padr a = match a with
  | Mem(r,i) -> preg r; pr "[%d]" i
  | Reg(r)   -> preg r
  in

  let disass adr =
    pr "%3d: " adr;
    begin match code.(adr) with
    | Load(a,d)   -> padr a; pr " <- %d" (match d with Int(i) -> i | Ptr(p) -> p)
    | Move(a,b)   -> padr a; pr " <- "; padr b
    | Push(a)     -> pr "push   "; padr a
    | Pop(a)      -> pr "pop    ";   padr a
    | Op(o,a,b,c) -> padr a; pr " <- "; padr b; pr " %s " o; padr c
    | Add(r,i)    -> pr "add    "; preg r; pr " %d" i
    | Call(a)     -> pr "call   "; padr a
    | Ret         -> pr "ret"
    | Jfalse(a,i) -> pr "jfalse "; padr a; pr " %d" i
    | Jptr(a,i)   -> pr "jptr   ";   padr a; pr " %d" i
    | Jmp(i)      -> pr "jmp    %d" i
    | Print(a)    -> pr "print "; padr a
    | Comm(t)     -> pr "comm %s" t
    | Halt        -> pr "halt"
    end;
    pr "\n"
  in

  for i=a1 to a2 do disass i done;
  flush ch
;;

#close "printf";;
