(*****************************************************************************
 *                                                                           *
 *                    Interface graphique pour le taquin                     *
 *                                                                           *
 * M. Quercia, le 20/10/96                                                   *
 *****************************************************************************)

#open "unix";;
let taquin_tk_path = "taquin.tk";; (* chemin d'accs au fichier TCL *)

(*
 * positions(table)
 *
 * Renvoie le vecteur des listes des postitions de chaque pice
 *)

let positions(t) =
  let n = vect_length(t) and p = vect_length(t.(0)) in
  let q = n*p in
  let pos = make_vect q [] in
  for i=0 to n-1 do for j=0 to p-1 do
    let x = t.(i).(j) in
    if x > 0 then pos.(x) <- (j,i) :: pos.(x)
  done done;
  pos
;;

(*
 * mouvement(dpart,arrive)
 *
 * "dpart", "arrive" = vecteurs de positions
 *
 * Dtermine si on passe de "dpart"  "arrive" par translation d'une pice.
 * Si oui, renvoie le triplet (pice,dx,dy),
 * sinon, lve l'exception Failure "mouvement illgal".
 *)

let mouvement(dpart,arrive) =

  (* cherche la premire diffrence *)
  let p = ref(1) in
  while !p < vect_length(dpart) & dpart.(!p) = arrive.(!p) do incr p done;
  if !p = vect_length(dpart) then failwith "mouvement illgal";

  (* vrifie que les pices suivantes n'ont pas boug *)
  for q = !p+1 to vect_length(dpart)-1 do
    if dpart.(q) <> arrive.(q) then failwith "mouvement illgal"
  done;

  (* vrifie que la pice "!p" est bien translate *)
  let (x,y) = hd(dpart.(!p)) and (x',y') = hd(arrive.(!p)) in
  let (dx,dy) = (x'-x,y'-y) in
  let l = map (fun (u,v) -> (u+dx,v+dy)) dpart.(!p) in
  if l <> arrive.(!p) then failwith "mouvement illgal";

  (* rsultat *)
  ([!p],dx,dy)
;;

(*
 * mouvements(liste)
 *
 * "liste" = liste d'chiquiers
 *
 * Vrifie que chaque mouvement est lgal et renvoie la liste des mouvements
 *)

let mouvements(liste) =
  let rec mvt(p,l) = match l with
  | [] -> []
  | t::suite -> let q = positions(t) in mouvement(p,q) :: mvt(q,suite)
  in
  mvt(positions(hd liste),tl liste)
;;


(*
 * comprime(liste)
 *
 * "liste" = liste de mouvements
 *
 * Renvoie une liste quivalente o les translations conscutives et gales
 * sont regroupes
 *)

let rec comprime(liste) = match liste with
| [] -> []
| [m] -> [m]
| (p,x,y)::([q],x',y')::suite when x'=x & y'=y & not(mem q p) -> comprime(((q::p),x,y)::suite)
| m::suite -> m :: (comprime suite)
;;

(*
 * anime combine liste
 *
 * "combime" = boolen disant s'il faut combiner les translations gales
 * "liste"   = liste d'chiquiers
 *
 * Animation graphique des mouvements
 *)

let anime combine liste =

  (* dmarre un interprteur *)
  let (chin,chout) = open_process "wish -name taquin" in
  output_string chout ("source " ^ taquin_tk_path ^ "\n");
  flush(chout);

  (* affiche l'chiquier initial *)
  let t = hd(liste) in
  let n = vect_length(t) and p = vect_length(t.(0)) in
  output_string chout "echiquier {";
  
  for i=0 to n-1 do
    output_string chout " {";
    for j=0 to p-1 do output_string chout (" " ^ string_of_int t.(i).(j)) done;
    output_char chout `}`
  done;
  output_string chout "}\n";
  flush(chout);

  (* constitue le vecteur des mouvements *)
  let mvts = mouvements liste in
  let mvts = vect_of_list(if combine then comprime mvts else mvts) in
  let i = ref 0 and sens = ref 1 and n = vect_length(mvts) in

  (* attend une action sur un bouton et agit en consquence *) 
  output_string chout "deplace 0 0 0\n"; flush(chout);
  let bouton = ref(input_line chin) in
  while !bouton <> "stop" do

    let ((p,dx,dy),j) = match 0 with
    | _ when !sens > 0 & !i < n -> (mvts.(!i),!i+1)
    | _ when !sens > 0          -> (([0],0,0),n)
    | _ when !sens < 0 & !i > 0 -> ((let (p,dx,dy) = mvts.(!i-1) in (p,-dx,-dy)),!i-1)
    | _                         -> (([0],0,0),0)
    in

    output_string chout "deplace {";
    do_list (fun x -> output_string chout (" " ^ (string_of_int x))) p;
    output_string chout ("} " ^ (string_of_int dx));
    output_string chout (" " ^ (string_of_int dy) ^ "\n");
    flush(chout);
    bouton := input_line(chin);

    i := j;
    if !bouton = "prec" then sens := -1;
    if !bouton = "suiv" then sens :=  1;

  done;

  (* arrte l'interprteur *)
  output_string chout "exit\n";
  flush(chout);
  let _ = close_process (chin,chout) in ()

;;
#close "unix";;
