(*****************************************************************************
 *                                                                           *
 *                               Jeu du taquin                               *
 *                                                                           *
 * M. Quercia, le 19/04/98                                                   *
 *****************************************************************************)

(* fonctions utilitaires fournies *)
include "tp-taquin.ml";;  (* interface graphique               *)
include "chrono.ml";;     (* chronomtrage                     *)
include "graphe.ml";;     (* algorithmes de parcours de graphe *)
include "memoire.ml";;    (* algorithmes de mmorisation       *)


                  (*+----------------------------+
                    |  Modlisation des taquins  |
                    +----------------------------+*)

(*
 * Un noeud du graphe est un couple (positions,table) o "positions" est un
 * vecteur donnant la position de chaque pice et "table" est la
 * matrice des cases. Les pices sont numrotes de 1  q, et le numro
 * 0 est rserv pour dsigner le trou.
 *)

type taquin = Taquin of ((int * int) vect) * (int vect vect);;

let print_taquin(Taquin(_,t)) = print_int_matrix(t);;
install_printer "print_taquin";;

let anime liste = anime false (map(fun (Taquin(_,t)) -> t) liste);;

(*
 * succ_taquin taquin
 *
 * Renvoie la liste des positions accessibles  partir de la position
 * fournie aprs dplacement d'une pice d'une case
 *)

let succ_taquin (Taquin(positions,table)) =

  let n = vect_length(table) and p = vect_length(table.(0)) in
  let liste = ref []          in (* liste rsultat *)
  let (x0,y0) = positions.(0) in

  (* effectue un dplacement et l'enregistre dans "!liste" *)
  let bouge(x1,y1) =
    let table' = copy_matrix(table) and pos' = copy_vect(positions) in
    let code = table.(y1).(x1)      in
    pos'.(0)    <- (x1,y1);
    pos'.(code) <- (x0,y0);
    table'.(y0).(x0) <- code;
    table'.(y1).(x1) <- 0;
    liste := Taquin(pos',table') :: !liste
  in

  if x0 > 0   then bouge(x0-1,y0);
  if y0 > 0   then bouge(x0,y0-1);
  if x0 < p-1 then bouge(x0+1,y0);
  if y0 < n-1 then bouge(x0,y0+1);

  !liste (* liste des successeurs *)
;;


(*
 * pions_mal_placs(n,n')
 *
 * Renvoie une minoration de la distance entre deux noeuds
 *)

let pions_mal_placs(Taquin(pos,_), Taquin(pos',_)) =
  let d = ref(0) in
  for i=1 to vect_length(pos)-1 do
    if pos.(i) <> pos'.(i) then incr d
  done;
  !d
;;

(*
 * distance de Manhattan
 *)

let manhattan(Taquin(pos,_), Taquin(pos',_)) =
  let d = ref(0) in
  for i=1 to vect_length(pos)-1 do
    let (x,y) = pos.(i) and (x',y') = pos'.(i) in
    d := !d + abs(x-x') + abs(y-y')
  done;
  !d
;;

(* constitue un noeud  partir d'un chiquier *)
let taquin_of_table(t) =

  let n = vect_length(t) and p = vect_length(t.(0)) in
  let q = n*p in
  let pos = make_vect q (0,0) in
  for i=0 to n-1 do for j=0 to p-1 do pos.(t.(i).(j)) <- (j,i) done done;
  Taquin(pos,t)
;;

                               (* +----------+
                                  |  Essais  |
                                  +----------+ *)


(*
 * Cherche un chemin de "source"  "but" avec les stratgies
 * de parcours "parcours" et de mmoire "mmoire".
 * Renvoie le chemin trouv, le temps de calcul et le nombre de noeuds visits.
 *)

let essai(source,but,parcours,p,mmoire) =

  let s_mem = match mmoire with
  | "liste" -> mmoire__liste
  | "table" -> mmoire__table
  | "arbre" -> mmoire__arbre
  | _       -> failwith "mmoire inconnue"
  and dist = match parcours with
  | "manh"  -> manhattan
  | "place" -> pions_mal_placs
  | _ -> (fun _ -> 0)
  in

  (* graphe direct *)
  let g = let (djavu,note,raz,taille) = s_mem() in
          {succ=succ_taquin;
           dist=dist;
           djavu=djavu;
           note=note;
           taille=taille}

  (* graphe rciproque pour larg2 *)
  and g'= let (djavu,note,raz,taille) = s_mem() in
          {succ=succ_taquin;
           dist=dist;
           djavu=djavu;
           note=note;
           taille=taille}
  in

  (* on y va ! *)
  let t = heure() in
  let (sol,taille) = match parcours with

    | "prof"  -> let sol = pprof(g,source,but,p) in (sol,g.taille())
    | "profv" -> let sol = pprof_variante(g,[[source],p],but) in (sol,g.taille())
	      
    | "larg"  -> g.note source 0;
                 let sol = plarg(g,[[source]],but) in (rev sol,g.taille())
	      
    | "larg2" -> g.note source 0; g'.note but 0;
                 let sol = plarg2(g,g',[[source]],[[but]],false) in
                 (sol,g.taille() + g'.taille())
	      
    | "manh"  -> let (sol,taille) = pheur(g,[([source],g.dist(source,but))],but)
                 in (rev sol,taille)

    | "place" -> let (sol,taille) = pheur(g,[([source],g.dist(source,but))],but)
                 in (rev sol,taille)

    | _       -> failwith "stratgie inconnue"

  in
  let t' = heure() in (sol,t'-.t,taille)
;;

(*************************** taquin 5x5 ************************************)
#open "printf";;
let but = taquin_of_table [| [| 1; 2; 3; 4; 5|];
                             [| 6; 7; 8; 9;10|];
                             [|11;12;13;14;15|];
                             [|16;17;18;19;20|];
                             [|21;22;23;24; 0|] |];;

let source = hd(random_chemin(succ_taquin,but,10));;

let job parcours mmoire =
  let (_,temps,taille) = essai(source,but,parcours,10,mmoire)
  in printf " %10d %6.2f" taille temps
in begin
  printf "              liste             table             arbre";
  do_list (fun p -> printf "\n%6s" p; flush stdout; do_list (job p)
          ["liste"; "table"; "arbre"])
          ["prof"; "larg"; "larg2"; "place"; "manh"];
  printf "\n";
  flush stdout
end;;

(* 30 mouvements, rsolution par Manhattan et animation graphique *)
begin
  let source = hd(random_chemin(succ_taquin,but,30)) in
  let (sol,temps,taille) = essai(source,but,"manh",0,"arbre") in
  printf " %10d %6.2f\n" taille temps; flush stdout;
  anime(sol)
end;;
