(*****************************************************************************
 *                                                                           *
 *                  Algorithmes de parcours de graphes                       *
 *                                                                           *
 * M. Quercia, le 19/04/98                                                   *
 *****************************************************************************)

type 'a graphe = {            (* graphe  noeuds de type 'a       *)
  succ   : 'a -> 'a list;     (* fonction de transition           *)
  dist   : 'a * 'a -> int;    (* distance estime entre 2 noeuds  *)
  djavu : 'a -> int;         (* dit si on est pass par un noeud *)
  note   : 'a -> int -> unit; (* note le passage par un noeud     *)
  taille : unit -> int        (* taille de la mmoire de passage  *)
};;


                   (* +----------------------------------+
                      |  Parcours en profondeur d'abord  |
                      +----------------------------------+ *)


(*
 * pprof(graphe,noeud,but,maxi)
 *
 * "graphe" = graphe  explorer
 * "noeud"  = point de dpart
 * "but"    = noeud  atteindre
 * "maxi"   = profondeur  ne pas dpasser
 *
 * Renvoie un chemin (probablement non optimal) "noeud" vers "but"
 * ou dclenche l'exception Failure "Echec".
 *)


let rec pprof(graphe,noeud,but,maxi) =
  if noeud = but then [noeud]
  else if maxi <= 0 then failwith "Echec"
  else begin
    graphe.note noeud maxi;
    noeud :: pprof_liste(graphe,graphe.succ(noeud),but,maxi-1)
  end

and pprof_liste(graphe,noeuds,but,maxi) = match noeuds with
| [] -> failwith "Echec"
| x::suite -> if graphe.djavu(x) >= maxi then pprof_liste(graphe,suite,but,maxi)
              else try pprof(graphe,x,but,maxi)
                   with Failure "Echec" -> pprof_liste(graphe,suite,but,maxi)
;;

(*
 * Remarque : pprof et pprof_liste sont mutuellement rcursives. 
 * Voici a titre d'exercice de style une version avec une seule fonction
 * rcursive : 
 *
 * pprof_variante(graphe,pchemins,but)
 *
 * "graphe"   = graphe  explorer
 * "pchemins" = liste de couples (chemin,profondeur maxi)  essayer
 * "but"      = noeud  atteindre
 *
 * Renvoie un chemin aboutissant au but s'il en existe, et lve l'exception
 * Failure "Echec" sinon
 *)

let rec pprof_variante(graphe,pchemins,but) = match pchemins with
| ((x::y) as ch,p)::suite ->
    if x = but then List.rev(ch)
    else if (p<=0) or (graphe.djavu(x)>=p) then pprof_variante(graphe,suite,but)
    else begin
      graphe.note x p;
      let suivants = List.map (fun t -> (t::ch,p-1)) (graphe.succ x)
      in pprof_variante(graphe,suivants @ suite, but)
   end
| _ -> failwith "Echec"
;;




                     (* +-------------------------------+
                        |  Parcours en largeur d'abord  |
                        +-------------------------------+ *)

(*
 * avance(graphe,chemins)
 *
 * "graphe" = graphe  explorer
 * "chemins"= liste des chemins dvelopps
 *
 * Rallonge chaque chemin d'une tape si possible et renvoie la
 * liste des chemins allongs.
 *)

let rec avance(graphe,chemins) = match chemins with
| [] -> []
| ch::suite -> let rec avance_un_chemin(liste) = match liste with
	       | [] -> []
	       | x::x' when graphe.djavu(x)>=0 -> avance_un_chemin(x')
               | x::x' -> graphe.note x 0; (x::ch) :: avance_un_chemin(x')
	       in
               let suivants = graphe.succ(List.hd ch) in
               avance_un_chemin(suivants) @ avance(graphe,suite)
	      
(* Variante de style : "avance_un_chemin" est maintenant locale  "avance", *)
(* ce qui conomise le passage des paramtres "graphe" et "ch".             *)
;;

(*
 * cherche(liste,ppt)
 *
 * "liste" = liste quelconque
 * "ppt"  = fonction disant si un lment de "liste" convient.
 *
 * Renvoie le premier lment de "liste" vrifiant "ppt", et lve l'exception
 * Failure "Non trouv" sinon.
 *)

let rec cherche(liste,ppt) = match liste with
| [] -> failwith "Non trouv"
| x :: suite -> if ppt(x) then x else cherche(suite,ppt)
;;

(*
 * plarg(graphe,chemins,but)
 *
 * "graphe" = graphe  explorer
 * "chemins"= liste des chemins dvelopps
 * "but"    = but  atteindre
 *
 * Renvoie un plus court chemin vers le but s'il y en a un.
 * S'il n'y en a pas, lve l'exception Failure "Echec" si les chemins ne
 * peuvent pas tre prolongs, et boucle si le graphe est infini.
 *
 * Attention : tous les noeuds des chemins doivent tre marqus comme dja vus
 *)

let rec plarg(graphe,chemins,but) =
  if chemins = [] then failwith "Echec"
  else try cherche(chemins,fun ch -> List.hd(ch) = but)
       with Failure "Non trouv" -> plarg(graphe,avance(graphe,chemins),but)

;;

              (* +---------------------------------------------+
                 |  Exploration en largeur par les deux bouts  |
                 +---------------------------------------------+ *)


(*
 * plarg2(g,g',sources,buts,inv)
 *
 * "g"       = graphe  explorer
 * "g'"      = graphe inverse
 * "sources" = liste de chemins issus d'une source
 * "buts"    = liste de chemins menant  un but
 * "inv"     = true s'il faut inverser le chemin trouv
 *
 * Renvoie un plus court chemin vers un point d'arrive s'il y en a un.
 * S'il n'y en a pas, lve l'exception Failure "Echec" si les chemins ne
 * peuvent pas tre prolongs, et boucle si le graphe est infini.
 *
 * Attention : tous les noeuds des chemins source doivent tre marqus comme
 * dja vus dans "g" et tous les noeuds des chemins but doivent tre marqus
 * comme dja vus dans "g'".
 *)

let rec plarg2(g,g',sources,buts,inv) =

  if sources = [] then failwith "Echec"
  else try let s = cherche(sources,fun ch -> g'.djavu(List.hd ch) >=0 ) in
           let b = cherche(buts,   fun ch -> List.hd(ch) = List.hd(s))   in
           match inv with
           | true  -> List.rev(b) @ List.tl(s)
           | false -> List.rev(s) @ List.tl(b)
       with Failure "Non trouv" -> plarg2(g',g,buts,avance(g,sources),not inv)
;;


                (* +----------------------------------------+
                   |  Exploration heuristiquement ordonne  |
                   +----------------------------------------+ *)


(*
 * pheur(graphe,vchemins,but)
 *
 * "graphe"  = graphe  explorer
 * "vchemins"= liste des chemins dvelopps avec cot estim
 * "but"     = noeud  atteindre
 *
 * Le cot d'un chemin est la somme de la longueur et de la distance
 * estime de l'extrmit au but.
 *
 * Renvoie un plus court chemin vers le but s'il y en a un et le nombre
 * de noeuds examins.
 * S'il n'y en a pas, lve l'exception Failure "Echec" si les chemins ne
 * peuvent pas tre prolongs, et boucle si le graphe est infini.
 *
 * Attention : seuls les noeuds internes des chemins doivent tre marqus
 * comme dja vus et la liste des chemins doit tre trie par cot croissant
 *)

let rec pheur(graphe,vchemins,but) = match vchemins with

| [] -> failwith "Echec"
| (ch,_)::suite ->
    if List.hd(ch) = but then (ch, graphe.taille() + List.length(vchemins))
    else begin
      graphe.note (List.hd ch) 0;

      (* prolonge "ch" d'une tape et met  jour la liste ordonne *)
      (* des chemins  explorer.                                   *)
      let suivants = ref(graphe.succ(List.hd ch))
      and liste = ref(suite) in
      while !suivants <> [] do let x = List.hd(!suivants) in

        suivants := List.tl(!suivants);
        if graphe.djavu(x) < 0 then begin
          let c = List.length(ch)+1+graphe.dist(x,but) in (* cot estim *)
          liste := insre_chemin(!liste,(x::ch,c))
        end

      done;
      (* ici, tous les prolongements de "ch" ont t ventuellement *)
      (* insrs dans !liste  la bonne place. On passe au meilleur *)
      (* chemin suivant.                                            *)
      pheur(graphe,!liste,but)
    end

and insre_chemin(liste,((ch,c) as v)) = match liste with
| (((x'::y'),c') as v')::suite ->
   if x' = List.hd(ch)
   then if c' <= c then liste else v::suite
   else if c' <= c then v' :: insre_chemin(suite,v)
   else v :: supprime_chemin(liste,List.hd ch)
| _ -> [v]

and supprime_chemin(liste,x) = match liste with
| (((x'::y'),c') as v')::suite ->
   if x' = x then suite else v' :: supprime_chemin(suite,x)
| _ -> []
;;

(*
 * Remarque : l'utilit de "supprime_chemin" n'est pas claire. Si l'on
 * laisse des chemins non performants dans la liste "vchemins" ils prennent
 * de la place, mais l'insertion d'un nouveau chemin se fait en temps quasi-
 * constant si l'heuristique est bonne (le nouveau chemin peut la plupart du
 * temps tre insr en dbut de liste).
 * Si l'on ne supprime pas les chemins non performants, alors il faut ajouter
 * un appel  djavu dans "pheur" pour vrifier qu'on ne traite pas deux
 * fois le mme noeud.
 *)


