(*+--------------------------------------------------------------------------+
  |                                                                          |
  |            Reconnaissance de tautologies, mthode de Davis et Putnam     |
  |                                                                          |
  +--------------------------------------------------------------------------+*)

(* Utilitaires sur la mise en arbre et l'affichage des expressions boolennes *)
include "tp-bool";;


           (* +--------------------------------------------------+
              |  Constitue la liste des variables d'une formule  |
              +--------------------------------------------------+ *)


let rec variables(f) = match f with
| Const(_)   -> []
| Var(x)     -> [x]
| Neg(u)     -> variables(u)
| Bin(_,u,v) -> fusion (variables u) (variables v)

and fusion u v = match (u,v) with
| [],_ -> v
| _,[] -> u
| x::u', y::v' -> if x < y then x :: (fusion u' v)
                  else if x > y then y :: (fusion u v')
                  else x :: (fusion u' v')
;;


                   (* +----------------------------------+
                      |  Substitution et simplification  |
                      +----------------------------------+ *)

(* ngation *)
let simplifie_ngation u = match u with
| Neg(v) -> v
| _ -> Neg(u)
;;

(* oprande gauche constant *)
let simplifie_gauche bin a v = match (opr(bin) a true, opr(bin) a false) with
| true,true   -> Const(true)
| false,false -> Const(false)
| true,false  -> v
| false,true  -> simplifie_ngation v
                 
(* oprande droit constant *)
and simplifie_droite bin u b = match (opr(bin) true b, opr(bin) false b) with
| true,true   -> Const(true)
| false,false -> Const(false)
| true,false  -> u
| false,true  -> simplifie_ngation u
;;

(* remplace Var(var) par Const(val) et simplifie *)
let rec simplifie f var val = match f with
| Const(_) -> f
| Var(v)   -> if v = var then Const(val) else f
| Neg(u) -> (match simplifie u var val with
             | Const(c) -> Const(not c)
             | Neg(v)   -> v
             | u'       -> Neg(u'))
| Bin(bin,u,v) -> (match (simplifie u var val ,simplifie v var val) with
             | Const(a), Const(b) -> Const(opr(bin) a b)
             | Const(a), v'  -> simplifie_gauche bin a v'
             | u', Const(b)  -> simplifie_droite bin u' b
             | u', v'        -> Bin(bin,u',v'))
;;
                 

                        (* +------------------------+
                           |  limine une variable  |
                           +------------------------+ *)


(* limine f var calcule les formules f1 = f/var=true
                                      f2 = f/var=false
   et les listes de variables v1,v2 de f1,f2
   et renvoie l'hexuplet (f1,v1,f2,v2,var,cot) o cot est
   le cot d'une vrification exhaustive de f1 et f2. *)

let limine f var =
  let f1 = simplifie f var true
  and f2 = simplifie f var false in
  let v1 = variables f1
  and v2 = variables f2 in
  let n1 = float_of_int(list_length v1)
  and n2 = float_of_int(list_length v2) in
  (f1,v1,f2,v2,var, 2.0**n1 +. 2.0**n2)
;;

(* limine_min f liste appelle limine pour chacune des variables de la liste
   et renvoie un hexuplet de cot minimum. *)

let rec limine_min f vars = match vars with
| []  -> failwith "cas impossible"
| [v] -> limine f v
| v::suite -> let (_,_,_,_,_,c)  as res  = limine f v
              and (_,_,_,_,_,c') as res' = limine_min f suite in
              if c <= c' then res else res'
;;


                       (* +--------------------------+
                          |  Cherche une rfutation  |
                          +--------------------------+ *)


(* rfute f variables cherche une rfutation de f. *)
(* opt = true s'il faut chercher le meilleur pivot *)
let rec refute opt f vars = match vars with
| []   -> if f = Const(false) then [] else failwith "tautologie"
| x::_ -> let (f1,v1,f2,v2,v,_) = if opt then limine_min f vars
                                         else limine f x
          in try (v,true) :: (refute opt f1 v1)
             with Failure "tautologie" -> (v,false) :: (refute opt f2 v2)
;;


                         (* +----------------------+
                            |  Mthode exhaustive  |
                            +----------------------+ *)

let rec exhaustif f vars valeurs = match vars with
| [] -> if value f valeurs = false then valeurs else failwith "tautologie"
| v::suite -> try exhaustif f suite ((v,true) :: valeurs)
              with Failure "tautologie" -> exhaustif f suite ((v,false) :: valeurs)

and value f valeurs = match f with
| Const(c) -> c
| Var(v)   -> assoc v valeurs
| Neg(u)   -> not(value u valeurs)
| Bin(bin,u,v) -> opr(bin) (value u valeurs) (value v valeurs)
;;


                                 (*+---------+
                                   |  essai  |
                                   +---------+*)

include "chrono.ml";;
#open "printf";;
begin
  let e = random_exp 10 16 in
  let v = variables(e)     in
  let f = Bin(Impl,e,e)    in
  printf "16 variables, hauteur 10\n"; flush stdout;

  printf "mthode exhaustive                      : "; flush stdout;
  let t0 = heure() in
  let _ = (try exhaustif f v [] with Failure "tautologie" -> [])
  in printf "%f\n" (heure() -. t0); flush stdout;

  printf "mthode de Davis et Putnam optimise     : "; flush stdout;
  let t0 = heure() in
  let _ = (try refute true f v  with Failure "tautologie" -> [])
  in printf "%f\n" (heure() -. t0); flush stdout;

  printf "mthode de Davis et Putnam non optimise : "; flush stdout;
  let t0 = heure() in
  let _ = (try refute false f v  with Failure "tautologie" -> [])
  in printf "%f\n" (heure() -. t0); flush stdout;

end;;
